import sys

import pyttman
from pyttman.clients.base import BaseClient
from pyttman.core.containers import Message, Reply, \
    ReplyStream


class CliClient(BaseClient):
    """
    Very simple CLI - client, offering a user interface
    with pyttman apps through the terminal window.
    """
    def run_client(self):
        print(f"\nPyttman v.{pyttman.__version__} - "
              f"CLI client", end="\n")
        try:
            print(f"{pyttman.settings.APP_NAME} is online! Start chatting"
                  f" with your app below."
                  "\n(?) Use Ctrl-Z or Ctrl-C plus Return to exit",
                  end="\n\n")
            while True:
                message = Message(input("[YOU]: "), client=self)
                reply = self.reply_to_message(message)
                if isinstance(reply, ReplyStream):
                    while reply.qsize():
                        print(f"[{pyttman.settings.APP_NAME.upper()}]: ",
                              reply.get().as_str())
                elif isinstance(reply, Reply):
                    print(f"{pyttman.settings.APP_NAME}:", reply.as_str())
                print()
        except (KeyboardInterrupt, EOFError):
            sys.exit(0)

    @staticmethod
    def publish(reply: Reply):
        print(reply.as_str())



import shutil
import subprocess
from pathlib import Path

from setuptools import setup
from twine.commands.upload import upload

# Replace these with your package information
package_name = "Pyttman"

# Get the package version dynamically
# the file location is in a sibling directory, add it to the path

# Upload to PyPI using twine
confirm = input("Deploying to PyPi.\n\n1: For pypi production, type 'production'"
                "\n2: For test.pypi.org, type 'test'\n\n")

if not Path("dist").exists():
    print("You need to build the package first. Run devtools/build.py.")
    exit(0)

print("Enter '__token__' for username, and the token for password")
if confirm == "production":
    subprocess.run(["twine", "upload", "dist/*"])
elif confirm == "test":
    subprocess.run(["twine", "upload", "--repository-url", "https://test.pypi.org/legacy/", "dist/*"])




import logging
import os
import shutil
import sys
import traceback
import warnings
from datetime import datetime
from importlib import import_module
from pathlib import Path

import requests

import pyttman
from pyttman.clients.builtin.cli import CliClient
from pyttman.core.ability import Ability
from pyttman.core.exceptions import PyttmanProjectInvalidException
from pyttman.core.internals import Settings, PyttmanApp, depr_raise
from pyttman.core.middleware.routing import AbstractMessageRouter
from pyttman_base_plugin import PyttmanPlugin

class TerraFormer:
    """
    Terraform a directory to start developing
    a Pyttman project by cloning it from GitHub.
    """

    def __init__(self, app_name: str, url: str):
        self.app_name = app_name
        self.url = url
        self.new_template_name = Path.cwd() / app_name
        self.template_file_name = Path("pyttman-template.zip")
        self.settings_file_path = self.new_template_name / Path("settings.py")
        self.template_dir_name = "pyttman-project-template-main"

    def terraform(self):
        """
        Clone the Pyttman template repository from GitHub
        and configure settings.py with the user provided app name.
        """
        old_template_name = Path.cwd() / self.template_dir_name
        scratch_file_name = Path.cwd() / self.template_file_name
        stream = requests.get(self.url, allow_redirects=True)

        with open(scratch_file_name, "wb") as f:
            f.write(stream.content)

        try:
            shutil.unpack_archive(scratch_file_name, Path.cwd())
            os.rename(old_template_name, self.new_template_name)
        except FileExistsError as e:
            if scratch_file_name.exists():
                os.remove(scratch_file_name)
            if old_template_name.exists():
                shutil.rmtree(old_template_name)
            raise FileExistsError("There is already an app with "
                                  "that name in this project.") from e

        with open(self.settings_file_path, "a", encoding="utf-8") \
                as settings_file:
            settings_file.write(f"\nAPP_NAME = \"{self.app_name}\"\n")

        os.remove(self.template_file_name)


def bootstrap_app(module: str = None, devmode: bool = False,
                  application_abspath: Path = None) -> PyttmanApp:
    """
    Bootstraps the framework with modules and configurations
    read from the settings.py found in the current path.

    Configuration for the logger, the clients used and
    Features are parsed, asserted and wrapped in Runner objects
    for the actual app starter to simply call 'run' on.

    :param module: Module in which the app source is located
    :param devmode: Provides only one runner with the CliClient in.
    :param application_abspath: Optional full path to the application
           directory
    """
    # This enables relative imports
    if application_abspath:
        sys.path.insert(0, Path(application_abspath).as_posix())
        os.chdir(application_abspath.as_posix())
    else:
        sys.path.insert(0, Path.cwd().as_posix())

    # First, find the settings.py. It should reside in the current directory
    # provided that the user is currently positioned in the app catalog for
    # their Pyttman project.
    if not (Path(module) / "settings.py").exists():
        raise PyttmanProjectInvalidException(f"\n\nApp '{module}' not found "
                                             f"in '{Path.cwd().as_posix()}'. "
                                             f"\nAre you running Pyttman in the "
                                             f"right directory?\nTip! If this happens "
                                             f"when unit testing, ensure the "
                                             f"correct 'app_name' is provided.")

    try:
        settings_module = import_module(f"{module}.settings")
        settings_names = [i for i in dir(settings_module)
                          if not i.startswith("_")]
        settings_config = {name: getattr(settings_module, name)
                           for name in settings_names}
        settings = Settings(**settings_config)
    except ImportError as e:
        print(traceback.format_exc())
        raise ImportError("An import error occurred when bootstrapping your "
                          "application. Check the modules imported in "
                          "'settings.py' and make sure they're installed and "
                          "spelled correctly.") from e

    if settings is not None:
        pyttman.settings = settings
        pyttman.is_configured = True
    else:
        raise NotImplementedError("Pyttman is improperly configured - "
                                  "settings module for app not found. ")

    # Configure the logger instance for pyttman.logger
    app_name = pyttman.settings.APP_NAME
    if pyttman.settings.APPEND_LOG_FILES:
        file_name = Path(f"{app_name}.log")
    else:
        file_name = Path(
            f"{app_name}-{datetime.now().strftime('%y%m%d-%H-%M-%S')}.log")

    log_file_dir = Path(pyttman.settings.LOG_FILE_DIR)
    if not log_file_dir.is_dir():
        os.mkdir(log_file_dir.as_posix())

    log_file_name = log_file_dir / file_name

    logging_handle = logging.FileHandler(
        filename=log_file_name, encoding="utf-8",
        mode="a+" if pyttman.settings.APPEND_LOG_FILES else "w")

    # Assign logging format - added in 1.1.9. Accounts for lower versions
    try:
        logging_format = pyttman.settings.LOG_FORMAT
    except AttributeError:
        logging_format = logging.BASIC_FORMAT

    logging_handle.setFormatter(logging.Formatter(logging_format))
    logger = logging.getLogger(app_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging_handle)

    if settings.LOG_TO_STDOUT:
        stdout_handle = logging.StreamHandler(sys.stdout)
        stdout_handle.setFormatter(logging.Formatter(logging_format))
        logger.addHandler(stdout_handle)

    # Set the configured instance of logger to the pyttman.PyttmanLogger object
    pyttman.logger.LOG_INSTANCE = logger

    # Import the router defined in ROUTER in settings.py
    try:
        message_router_config = settings.ROUTER.get(
            "ROUTER_CLASS"
        ).split(".")
    except AttributeError:
        depr_raise("Failed to load ROUTER config; 'ROUTER' is a required "
                   "attribute in settings.py.",
                   "1.3.2")

    # Load plugins
    pyttman.logger.log("- [Pyttman]: Loading plugins...")
    try:
        plugins: list[PyttmanPlugin] = settings.PLUGINS
    except AttributeError as e:
        raise AttributeError("PLUGINS is missing from settings.py.") from e
    pyttman.logger.log(f"- [Pyttman]: Plugins loaded: {[i.__class__.__name__ for i in plugins]}")

    message_router_class_name = message_router_config.pop()
    message_router_module = ".".join(message_router_config)
    message_router_module = import_module(message_router_module)

    if not (message_router_class := getattr(message_router_module,
                                            message_router_class_name)):
        raise ImportError(f"Pyttman could not find the router "
                          f"'{message_router_class_name} in "
                          f"{message_router_class_name}. "
                          f"Verify the MESSAGE_ROUTER setting in settings.py.")

    # Retrieve the help keyword from settings
    if not (help_keyword := settings.ROUTER.get("HELP_KEYWORD")):
        raise AttributeError("'HELP_KEYWORD' not defined in settings.py. "
                             "Please define a word for the automatic "
                             "help page generation to trigger on in your "
                             "app whenever it is present as the first "
                             "occurring string in a message.\nExample: "
                             "'HELP_KEYWORD' = 'help'")

    # Retrieve command-unknown-responses from settings
    if not (command_unknown_responses := settings.ROUTER.
            get("COMMAND_UNKNOWN_RESPONSES")):
        raise ValueError("There are no responses provided for when "
                         "no intents match a query. Define these in "
                         "ROUTER['COMMAND_UNKNOWN_RESPONSES'] as "
                         "a list of strings")

    # Import the client classes defined in CLIENTS in settings.py
    if not len(settings.CLIENT) and not devmode:
        raise ValueError("A Client is required for Pyttman to "
                         "start your app in Client mode. Define a Client "
                         "in settings.py. Refer to the documentation for "
                         "examples.")

    # Instantiate router and provide the APP_NAME from settings
    message_router: AbstractMessageRouter = message_router_class(
        abilities=None,
        intent_unknown_responses=command_unknown_responses,
        help_keyword=help_keyword)

    # If devmode is active, return only one CliClient in a runner.
    if devmode:
        client = CliClient(message_router=message_router,
                           plugins=plugins)
        app = PyttmanApp(client=client,
                         name=settings.APP_NAME,
                         settings=settings,
                         plugins=plugins)
        pyttman.app = app
        app.abilities = load_abilities(settings)
        message_router.abilities = app.abilities
        message_router.app = app
        prepare_app(module)
        del settings.CLIENT
        return app

    # Start the client
    if not isinstance(settings.CLIENT, dict):
        raise TypeError("The CLIENT config must be a single dict, "
                        "containing the full reference to the client "
                        "class")
    try:
        client_class_config = settings.CLIENT.pop("class").split(".")
    except Exception:
        raise ValueError("The dictionary containing Client configuration "
                         "must contain the key 'class', which should be "
                         "the import path for the Client class. Verify "
                         "the 'CLIENT' property in settings.py. See the "
                         "Client documentation for more help on this subject.")
    try:
        client_class_name = client_class_config.pop()
        client_module_name = ".".join(client_class_config)
        client_module = import_module(client_module_name)

        # Provide the client with message_router and runner with client
        client_class = getattr(client_module, client_class_name)
    except Exception as e:
        raise RuntimeError(f"Cannot start client. Check "
                           f"the following error and correct "
                           f"any syntax error in settings.py: "
                           f"\n{e}") from e
    client = client_class(message_router=message_router,
                          **settings.CLIENT,
                          plugins=plugins)

    # Create a log entry for app start
    pyttman.logger.log(f"- [Pyttman]: App {app_name} started: {datetime.now()} -- ")
    app = PyttmanApp(client=client,
                     name=settings.APP_NAME,
                     settings=settings,
                     plugins=plugins)

    pyttman.app = app
    app.abilities = load_abilities(settings)
    prepare_app(module)
    message_router.abilities = app.abilities
    message_router.app = app
    del settings.CLIENT
    return app


def prepare_app(module):
    try:
        import_module(f"{module}.app")
    except ImportError:
        # The app is missing this file; no problemo.
        pass


def load_abilities(settings: Settings) -> set:
    # Set the abilities of the router to the abilities from settings.py
    ability_objects_set = set()
    for ability in settings.ABILITIES:
        assert not isinstance(ability, Ability), \
            f"The ability '{ability}' is instantiated. Please redefine " \
            f"this ability as only the reference to the class, " \
            f"as in 'MyClass', not 'MyClass()'. "

        ability_module_config = ability.split(".")
        ability_class_name = ability_module_config.pop()
        ability_module_name = ".".join(ability_module_config)
        ability_module = import_module(ability_module_name)
        ability_class = getattr(ability_module, ability_class_name)
        ability_objects_set.add(ability_class())

    if not len(ability_objects_set):
        warnings.warn("There are no ability classes loaded - "
                      "your app will not have any replies.")
    return ability_objects_set



from pyttman.core.internals import PyttmanApp

from pyttman import version
from pyttman.tools.logger.logger import PyttmanLogger

__version__ = version.__version__
__author__ = "Simon Olofsson"
__copyright__ = "(c) Pyttman development Team 2020-2022"
__licence__ = "MIT"


class _SettingsNotConfigured:
    def __getattr__(self, item):
        raise NotImplementedError("pyttman.settings cannot be "
                                  "accessed in this scope. "
                                  "To use pyttman.settings  "
                                  "you must use the Pyttman CLI."
                                  "If you're looking for an "
                                  "interactive shell for debugging, "
                                  "use 'pyttman shell <app name>'. ")


class _AppNotConfigured:
    def __getattr__(self, item):
        raise NotImplementedError("pyttman.app was accessed before its "
                                  "creation. This error can occur if you're "
                                  "trying to decorate functions as lifecycle "
                                  "hooks, in an ability -or intent module, "
                                  "which is included in the 'ABILITIES' list "
                                  "in settings.py. To mitigate this error, "
                                  "move lifecycle hooks to a separate module.")


app: PyttmanApp | None = None
settings = _SettingsNotConfigured
is_configured = False

logger = PyttmanLogger()
"""
Logs function return value and/or exceptions to the application 
log file. 
"""


"""
I love you

- That was written by my girlfriend without me knowing about it,
on that very line. I left my computer unlocked some time during
a weekend. It's so sweet that I can't remove it, so it's here,
a part of the framework forever, that's the way it is. Consider
yourself the founder of an easter egg.

// Simon Olofsson, lead developer and founder of Pyttman
"""



import abc
import random
import warnings
from copy import copy
from typing import List, Any, Iterable

import pyttman
from pyttman.core.exceptions import PyttmanProjectInvalidException
from pyttman.core.entity_parsing.parsers import parse_entities
from pyttman.core.ability import Ability
from pyttman.core.intent import Intent
from pyttman.core.containers import MessageMixin, Reply, ReplyStream, Message
from pyttman.core.internals import _generate_error_entry, PyttmanApp


class AbstractMessageRouter(abc.ABC):
    """
    Abstract class for a MessageRouter.

    MessageRouters delegate Message objects
    coming from Clients, originating from some
    kind of platform with a user interface.

    Message routers can be very simple and
    linearly search through Abilities and their
    Intent classes and select the first one
    matching - or they can be powered by internal
    caches, learn patterns and be powered by
    Machine learning.

    Users should rarely encounter this class as
    it's being used outside the scope of apps
    developed in Pyttman.
    """

    help_keyword = "help"

    def __init__(self, abilities: List[Ability],
                 intent_unknown_responses: List[str],
                 help_keyword: str,
                 app: PyttmanApp = None,
                 **kwargs):
        self.app = app
        self.abilities = abilities
        self.help_keyword = help_keyword
        self.intent_unknown_responses = intent_unknown_responses
        [setattr(self, k, v) for k, v in kwargs.items()]

    @abc.abstractmethod
    def get_matching_intent(self, message: MessageMixin) -> List[Intent]:
        """
        Return all matching Intent classes which match
        a given message, as made evident by their
        lead and trail configuration.

        :param message: MessageMixin subclassed object, from client
        :return: collection of Intents, if multiple.
        """
        pass

    @abc.abstractmethod
    def get_reply(self, message: MessageMixin) -> Reply:
        """
        Return the Reply from matching Intent.
        If no command matched, return a response from
        default_responses in settings.

        If the command contains help, return the help
        string for the matching command.
        :param message: MessageMixin subclassed object, from client
        """
        pass

    def process(self,
                message: Message,
                intent: Intent,
                keep_alive_on_exc=True) -> Reply | ReplyStream:
        """
        Iterate over all EntityFieldValueParser objects and the name
        of the field it's allocated as.

        The strings present in 'lead' and 'trail' in the Intent are
        filtered out as for them not to be parsed by the Entity parser.

        :param intent: The Intent class chosen to provide a Reply to the user.
        :param message: MessageMixin object
        :param keep_alive_on_exc: Keeps the main loop running if exceptions
        occur in the application logic, and replies with an error message
        fetched from the application settings. Defaults to True.
        :return: Reply, logic defined in the 'respond' method
        """
        joined_patterns = set()

        if intent.exclude_lead_in_entities is True:
            joined_patterns.update(intent.lead)
        if intent.exclude_trail_in_entities is True:
            joined_patterns.update(intent.trail)
        truncated_content = [i for i in message.content
                             if i.casefold() not in joined_patterns]
        truncated_message = Message(content=truncated_content)
        try:
            entities: dict[str: Any] = parse_entities(
                message=truncated_message,
                entity_fields=intent.user_entity_fields,
                original_message_content=copy(message.content),
                exclude=intent.ignore_in_entities)

            message.entities = {k: v.value for k, v in entities.items()}
        except Exception as e:
            reply = _generate_error_entry(message, e)
            if keep_alive_on_exc is False:
                raise e
            return reply

        try:
            intent.before_respond(message)

            if self.app is not None:
                message = self.app.execute_plugins_before_intent(message)

            reply = intent.respond(message=message)

            if self.app is not None:
                reply = self.app.execute_plugins_after_intent(reply)

            intent.after_respond(message, reply)
        except Exception as e:
            reply = _generate_error_entry(message, e)
            if keep_alive_on_exc is False:
                raise e

        original_reply = copy(reply)
        try:
            if not any((isinstance(reply, Reply), isinstance(reply, ReplyStream))):
                if isinstance(reply, Iterable) and not isinstance(reply, str):
                    reply = Reply(reply)
                else:
                    reply = ReplyStream(reply)
        except Exception:
            raise PyttmanProjectInvalidException(
                f"Could not return reply from intent '{intent}' due to "
                f"a misconfiguration of the response type. '{intent.respond}' "
                f"returned: '{original_reply}'")

        constraints = {
            bool(reply is not None),
            bool(isinstance(reply, Reply) or isinstance(reply, ReplyStream))
        }

        if False in constraints:
            raise ValueError(f"Improperly configured Intent class: "
                             f"{intent.__class__.__name__}."
                             f"respond method returned '{type(reply)}', "
                             f"expected Reply or ReplyStream")

        for entity_field in intent.user_entity_fields.values():
            entity_field.reset()

        return reply


class FirstMatchingRouter(AbstractMessageRouter):
    """
    Iterates over intents linearly.
    No calculation performed when routing messages and
    multiple abilities matches a Message - the first one
    in order is chosen.
    """
    def get_reply(self, message: Message) -> Reply:
        if self.app is not None:
            message = self.app.execute_plugins_before_router(message)

        try:
            if not (matching_intents := self.get_matching_intent(message)):
                original_message = copy(message.content)

                if self.app is not None:
                    reply = self.app.execute_plugins_no_intent_match(message)
                    if reply.content != original_message:
                        return reply
                return Reply(random.choice(self.intent_unknown_responses))
        except Exception as e:
            return _generate_error_entry(message, e)

        if len(matching_intents) > 1:
            warning_message = "More than one Intent matched a message. " \
                              "Consider changing the Router class in " \
                              "the settings module for this project if " \
                              "you wish your users to receive a choice " \
                              "of which command to execute in situations " \
                              f"like these. Matching intents: " \
                              f"{matching_intents}"
            warnings.warn(warning_message)
            pyttman.logger.log(warning_message)

        # Take the first matching one and use it to reply to the Message.
        chosen_intent = matching_intents.pop()

        # Return the auto-generated help segment for the Intent if the
        # HELP keyword is the first occurring word in the message.
        try:
            first_word = message.sanitized_content(preserve_case=False)[0]
        except IndexError:
            pass
        else:
            if first_word == self.help_keyword.lower().strip():
                if chosen_intent is not None:
                    return Reply(chosen_intent.generate_help())
                # else:
                #  TODO - Return help chapter for ability
        try:
            reply = self.process(message=message, intent=chosen_intent)
        except Exception as e:
            reply: Reply = _generate_error_entry(message, e)
        return reply

    def get_matching_intent(self, message: MessageMixin) -> List[Intent]:
        """
        Perform a linear search over intents for abilities.
        The matching one first in the sequence is chosen to
        reply the user.

        If more than one Intent would match, the user is notified
        with a warning as to investigate the design of their Intent
        scheme. It may be wiser to use another MessageRouter class
        which supports multiple match routing.
        :param message:
        :return: List of Intent instances which match the intent
        """
        matching_intents = []
        for ability in self.abilities:
            for intent in ability.intents:
                try:
                    intent = intent(storage=ability.storage,
                                    ability=ability)
                    if intent.matches(message):
                        matching_intents.append(intent)
                except TypeError as e:
                    raise TypeError(f"The intent {intent} did not behave"
                                    f" as expected - see full traceback.") \
                        from e
        return matching_intents



from datetime import datetime

import discord

from pyttman import logger
from pyttman.clients.base import BaseClient
from pyttman.clients.community.discord.misc import DiscordMessage
from pyttman.core.containers import ReplyStream, Reply
from pyttman.core.exceptions import ClientImproperlyConfiguredError
from pyttman.core.internals import _generate_error_entry
from pyttman.core.middleware.routing import AbstractMessageRouter


class DiscordClient(discord.Client, BaseClient):
    """
    Client class for the Discord chat service.
    Token and guild are both required.

    It implements the 'on_message' method overloaded
    from `discord.Client.on_message` , and plugging it
    in to the Pyttman ecosystem, allowing the application
    logic to take care of responding to the message.

    Since the `discord.Client` class provides an
    impressive array of functionality, this class
    can be subclassed by users who wish to implement
    behavior for their application upon any of the
    other available event triggers from the discord.Client
    class, available to read in their documentation.

    Below are attributes and methods unique to the
    `DiscordClient` class.

    Attributes
    -----------
    token: :class:`str`
        Token used when authenticating with Discord's API.
        The token is deleted from the instance at time of
        connection.
    guild: :class:´str`
        Guild ID from Discord, in which the Bot is registered
        to and is allowed to connect to.
    message_startswith: :class:`str`
        Optional string, setting a condition for all messages
        being required to start with said string, for the bot
        to consider the message as a command to parse.
        Defaults to empty string, allowing all messages to
        be parsed.
    message_endswith: :class:`str`
        Optional string, setting a condition for all messages
        being required to end with said string, for the bot
        to consider the message as a command to parse.
        Defaults to empty string, allowing all messages to
        be parsed.
    """
    token: str = ""
    guild: int = None
    message_startswith: str = ""
    message_endswith: str = ""
    message_length_limit = 2000

    def __init__(self,
                 message_router: AbstractMessageRouter,
                 token: str,
                 guild: int,
                 *args,
                 **kwargs):
        try:
            discord_intent_flags = kwargs["discord_intent_flags"]
            discord_intents = discord.Intents(**discord_intent_flags)
            if not discord_intents:
                raise ClientImproperlyConfiguredError(
                    "Pyttman cannot start the Discord client without the "
                    "discord intent 'message_content' being set to True, "
                    "since the contents in the message are required to "
                    "do the routing and deliver the Entities found in the "
                    "message.")
        except KeyError:
            raise ClientImproperlyConfiguredError(
                "Could not find 'discord_intent_flags' in 'CLIENT' section "
                "in settings.py for this app, and it's required. The app cannot "
                "start. Refer to the documentation for settings.py to setup "
                "the client and the discord intent flags correctly.")

        message_startswith = kwargs.get("message_startswith")
        message_endswith = kwargs.get("message_endswith")
        self.message_startswith = message_startswith or self.message_startswith
        self.message_startswith = message_endswith or self.message_endswith
        self._token = token
        self.guild = guild
        BaseClient.__init__(self,
                            message_router,
                            **kwargs)
        discord.Client.__init__(self,
                                token=token,
                                guild=guild,
                                intents=discord_intents,
                                **kwargs)

    async def on_ready(self):
        logger.log(f"- [DiscordClient]: App online on discord.")

    async def send_message_segmented(self,
                                     outgoing_message: Reply,
                                     channel: any) -> None:
        for segment in outgoing_message.segmented(self.message_length_limit):
            await channel.send(segment)

    async def on_message(self, message: DiscordMessage) -> None:
        """
        Overloads on_message in discord.Client().The
        method is called implicitly whenever an incoming
        message over the discord interface is intercepted.
        All messages are parsed, thus it's a good idea to
        verify that the author is not self to prevent endless
        loops.
        :param message: discord.Message
        :return: None
        """
        if message.author == self.user:
            return

        # Add the client to the attributes, passed on to the message object
        attrs = {"created": datetime.now(),
                 "client": self,
                 "channel": message.channel}

        for attr_name in message.__slots__:
            try:
                attrs[attr_name] = getattr(message, attr_name)
            except AttributeError:
                # It was likely a method - skip it
                continue

        # Accessing the protected _state field is unfortunately necessary here
        discord_message = DiscordMessage(state=attrs.get("_state"),
                                         data=attrs, **attrs)

        msg_as_str: str = discord_message.as_str()
        if self.message_startswith and not msg_as_str.startswith(
                self.message_startswith) \
                or self.message_endswith and not msg_as_str.endswith(
            self.message_endswith):
            return

        async with message.channel.typing():
            try:
                reply = self.reply_to_message(discord_message)
                if isinstance(reply, ReplyStream):
                    while reply.qsize():
                        await self.send_message_segmented(reply.get(), message.channel)
                else:
                    await self.send_message_segmented(reply, message.channel)
            except Exception as e:
                await discord_message.channel.send(
                    _generate_error_entry(discord_message, e).as_str())

    def run_client(self):
        if not self._token:
            raise ClientImproperlyConfiguredError(
                "Cannot connect - missing authentication."
                "'token' must be provided when "
                "using the Discord client. Specify these "
                "values in the dictionary for the client "
                "as {\"token\": \"foo_token\", \"guild\": "
                "\"foo_guild\"}")

        _token = self._token
        del self._token

        try:
            self.run(_token)
        except Exception as e:
            raise ClientImproperlyConfiguredError(
                "DiscordClient ran in to a problem. "
                "Please ensure that all data is provided "
                "correctly in settings.py."
                f"Exact error message: \"{e}\"")



import shutil
import subprocess
import sys
from pathlib import Path
from time import sleep

# Set the current working dir to parent directory
sys.path.append(Path.cwd().parent.as_posix())

CATALOGS = ("build", "dist")
BUILD_CMD = "python -m setup sdist bdist_wheel".split()


if __name__ == "__main__":
    [shutil.rmtree(i) for i in CATALOGS if Path(i).exists()]
    sleep(1)
    subprocess.run(BUILD_CMD)



from pyttman.version import __version__
from pathlib import Path
from setuptools import setup, find_packages


HERE = Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


# This call to setup() does all the work
setup(
    name="Pyttman",
    version=__version__,
    description="The Python chatbot framework with batteries included",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Hashmap-Software-Agency/Pyttman",
    author="Simon Olofsson",
    project_urls={
        "Bug Tracker": "https://github.com/dotchetter/Pyttman/issues",
        "Changelog": "https://github.com/dotchetter/Pyttman/blob/main/CHANGELOG.md"
    },
    author_email="simon@hashmap.se",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    packages=find_packages(exclude=["*.tests", "*.tests.*",
                                    "tests.*", "tests"]),
    include_package_data=True,
    install_requires=[
        "multidict",
        "discord.py",
        "requests",
        "py7zr",
        "ordered_set",
        "pyttman-plugin-base",
    ],
    entry_points={
        "console_scripts": [
            "pyttman=pyttman.tools.pyttmancli:run",
        ]
    },
)
